#include <Arduino.h>

/*
 Udp NTP Client

 Про NTP протокол читать http://en.wikipedia.org/wiki/Network_Time_Protocol

 NTP серверы искать http://tf.nist.gov/tf-cgi/servers.cgi

 */

#include <SPI.h>
#include <Ethernet.h>
#include <EthernetUdp.h>
#include <stdint.h>

#define pinRelay 7                              // Сюда включим реле

#define timeIntervalToNextAttempt 600000        // Интервал обычной проверки NTP 10 минут
#define timeIntervalIfFirstAttemptFall  300000  // Интервал повторной проверки после перезагрузки 5 минут
#define timeIntervalIfSecondAttemptFall 3600000 // Интервал если перезагрузка не помогла - один час.

// Проверка работоспособности:
//#define timeIntervalToNextAttempt 60000        // Интервал обычной проверки NTP 1 минута
//#define timeIntervalIfFirstAttemptFall  60000  // Интервал повторной проверки после перезагрузки 1 минута
//#define timeIntervalIfSecondAttemptFall 120000 // Интервал если перезагрузка не помогла 2 минуты.


#define timeZone 4	                            // Временная зона, где мы живем

byte mac[] = { 0x90, 0xA2, 0xDA, 0x0F, 0x79, 0xF0 }; // MAC address для шилда.
//IPAddress ip = ( 192, 168, 1, 177 );

unsigned int localPort = 8888;                  // local port для прослушки UDP пакетов
IPAddress timeServer(132, 163, 4, 101);         // сервер NTP time-a.timefreq.bldrdoc.gov
IPAddress timeServer02(128, 138, 140, 44);      // сервер NTP utcnist.colorado.edu
IPAddress timeServer03(216, 228, 192, 69);      // сервер NTP nist-time-server.eoni.com
const int NTP_PACKET_SIZE= 48;                  // NTP time stamp - первые 48 байтов пакета
byte packetBuffer[ NTP_PACKET_SIZE];            // Буфер для входящих и исходящих пакетов
EthernetUDP Udp;                                // Создаем UDP
const unsigned long seventyYears = 2208988800UL;// Время Unix началось 1 января 1970 года. В секундах это 2208988800

unsigned long timeWhenCheckNTP ;                // Время, когда пора проверить NTP сервер.
// unsigned long timeWhenGotNTP;
String timeNowIs = "";                          // Здесь храним запись текущего времени;

uint8_t attemptCounter = 0;                     // Считаем попытки достучаться до вервера;
bool isNeedToReboot = false;                    // Надо ли перегружать роутер?
// Объявляем функции:
unsigned long sendNTPpacket(IPAddress& address);    // Функция отправки запроса NTP
void fUDPReadAndGotTime(void);                  // Функция разбора ответа NTP


void setup()
{
    pinMode(pinRelay, OUTPUT);
    digitalWrite(pinRelay, LOW);
    Serial.begin(9600);
    //while (!Serial) {; }
//    Ethernet.begin(mac, ip);

    while (Ethernet.begin(mac) == 0) {                                      // start Ethernet and UDP
        Serial.println("Failed to configure Ethernet using DHCP");
        delay(1000);
    }
    Udp.begin(localPort);
    delay(5000);
    sendNTPpacket(timeServer);
    fUDPReadAndGotTime();
    timeWhenCheckNTP = millis() + timeIntervalToNextAttempt;
}

void loop() {

/***************************** Проверяем NTP, три сервера **************************/
    if (millis()> timeWhenCheckNTP) {
        sendNTPpacket(timeServer);                          // Отправляем запрос
        delay (1000);
        if (Udp.parsePacket()) {                            // Если есть ответ
            fUDPReadAndGotTime();                           // Читаем время
            attemptCounter = 0;                             // Обнуляем счетчик перезагрузок
        }
        else {
            sendNTPpacket(timeServer02);
            delay(1000);
            if (Udp.parsePacket()) {
                fUDPReadAndGotTime();
                attemptCounter = 0;
            }
            else {
                sendNTPpacket(timeServer02);
                delay(1000);
                if (Udp.parsePacket()) {
                    fUDPReadAndGotTime();
                    attemptCounter = 0;
                }
                else {                                      // Если ни один из серверов не ответил
                    isNeedToReboot = true;                  // Установка флага перезагрузки
                    attemptCounter +=1;                     // Увеличение счетчика перезагрузок
                }
            }
        }
/***************************** Определяем время следующей перезагрузки **************************/
        if (attemptCounter == 0) {                          // Если все хорошо то следующая проверка через 10 минутЖ;
            timeWhenCheckNTP = millis() + timeIntervalToNextAttempt;
        }
        if (attemptCounter == 1) {                          // Если перегрузились - проверяем через пять минут
            timeWhenCheckNTP = millis() + timeIntervalIfFirstAttemptFall;
        }
        if (attemptCounter > 1) {                           // Если перегрузка не помогла - перегружаемся каждый час
            timeWhenCheckNTP = millis() + timeIntervalIfSecondAttemptFall;
        }

/******************************* Перегружаем, если надо ****************************************/
        if(attemptCounter < 50) {                           // Перегружаемся не более двух суток
            if (isNeedToReboot) {                           // Если флаг перезагрузки установлен
                isNeedToReboot = false;                     // Сбрасываем его
                Serial.println("Relay!! Relay!! Relay!!");  // Оповещаем в сериал-порт
                digitalWrite(pinRelay, HIGH);               // Включаем реле
                delay(30000);                               // Ждем 30 секунд
                digitalWrite(pinRelay, LOW);                // Отключаем реле
            }
        }
    }
}

/********************************** Отправка пактета на сервер NTP ***************************/
unsigned long sendNTPpacket(IPAddress& address) {
    memset(packetBuffer, 0, NTP_PACKET_SIZE);                               // set all bytes in the buffer to 0
                                                                            // Initialize values needed to form NTP request
                                                                            // (see URL above for details on the packets)
    packetBuffer[0] = 0b11100011;                                           // LI, Version, Mode
    packetBuffer[1] = 0;                                                    // Stratum, or type of clock
    packetBuffer[2] = 6;                                                    // Polling Interval
    packetBuffer[3] = 0xEC;                                                 // Peer Clock Precision
                                                                            // 8 bytes of zero for Root Delay & Root Dispersion
    packetBuffer[12]  = 49;
    packetBuffer[13]  = 0x4E;
    packetBuffer[14]  = 49;
    packetBuffer[15]  = 52;
                                                                            // all NTP fields have been given values, now
                                                                            // you can send a packet requesting a timestamp:
    Udp.beginPacket(address, 123);                                          //NTP requests are to port 123
    Udp.write(packetBuffer,NTP_PACKET_SIZE);
    Udp.endPacket();
}
/******************************** Анализ пактета от NTP **************************************/
void fUDPReadAndGotTime(void) {
    Udp.read(packetBuffer,NTP_PACKET_SIZE);                             // read the packet into the buffer
    unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);  //the timestamp starts at byte 40 of the received packet and is four bytes,
    unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);   // or two words, long. First, esxtract the two words:
    unsigned long secsSince1900 = highWord << 16 | lowWord;             // combine the four bytes (two words) into a long integer
    Serial.print("Seconds since Jan 1 1900 = " );                       // this is NTP time (seconds since Jan 1 1900)
    Serial.println(secsSince1900);
    Serial.print("Unix time = ");                                       // now convert NTP time into everyday time:
    unsigned long epoch = secsSince1900 - seventyYears + (timeZone*3600); // subtract seventy years + Time Zone correction:
    Serial.println(epoch);                                              // print Unix time:
    Serial.print("The UTC time is ");                                   // UTC is the time at Greenwich Meridian (GMT)
    Serial.print((epoch  % 86400L) / 3600);                             // print the hour (86400 equals secs per day)
        timeNowIs = (String) ((epoch  % 86400L) / 3600);
    Serial.print(':');
        timeNowIs += ":";
    if ( ((epoch % 3600) / 60) < 10 ) {                                 // In the first 10 minutes of each hour, we'll want a leading '0'
        Serial.print('0');
        timeNowIs += "0";
    }
    Serial.print((epoch  % 3600) / 60);                                 // print the minute (3600 equals secs per minute)
        timeNowIs += (String) ((epoch  % 3600) / 60);
    Serial.print(':');
        timeNowIs += ":";
    if ( (epoch % 60) < 10 ) {                                          // In the first 10 seconds of each minute, we'll want a leading '0'
        Serial.print('0');
        timeNowIs += "0";
    }
    Serial.println(epoch %60);                                          // print the second
        timeNowIs += (String) (epoch %60);
        Serial.println(timeNowIs);
}
